import numpy as np
from profiling import profiling_data
from profiling import tools

from autolens import exc
from autolens.profiles import light_profiles


class Convolver(object):

    def __init__(self, mask, blurring_mask, kernel):
        """
        Class to create number array and frames used to convolve_image a psf with a 1D vector of non-masked values.

        Parameters
        ----------
        mask : Mask
            A mask where True eliminates data.
        mask : Mask
            A mask of pixels outside the mask but whose light blurs into it after convolution.
        kernel : masked_image.PSF or ndarray
            An array representing a PSF psf.

        Attributes
        ----------
        blurring_frame_indexes: [ndarray]
            An array of frames created by the frame maker. Maps positions in the psf to values in the 1D vector for
            masked pixels.
        image_frame_indexes: [ndarray]
            An array of frames created by the frame maker. A frame maps positions in the psf to values in the 1D
            vector.
        """

        if kernel.shape[0] % 2 == 0 or kernel.shape[1] % 2 == 0:
            raise exc.KernelException("Kernel must be odd")
        if mask.shape != blurring_mask.shape:
            raise exc.KernelException("Mask and Blurring mask must be same shape to generate Convolver")

        self.mask = mask
        self.kernel = kernel
        self.kernel_shape = kernel.shape
        self.mask_index_array = np.full(self.mask.shape, -1)

        count = 0

        for x in range(self.mask.shape[0]):
            for y in range(self.mask.shape[1]):
                if not self.mask[x, y]:
                    self.mask_index_array[x, y] = count
                    count += 1

        self.frame_array = []
        self.frame_kernel_array = []
        for x in range(self.mask_index_array.shape[0]):
            for y in range(self.mask_index_array.shape[1]):
                if not self.mask[x][y]:
                    frame, kernel_frame = self.frame_at_coords((x, y))
                    self.frame_array.append(frame)
                    self.frame_kernel_array.append(kernel_frame)

        self.frame_lengths = np.asarray(list(map(lambda frame: frame.shape[0], self.frame_array)), dtype='int')

        self.blurring_frame_array = []
        self.blurring_frame_kernel_array = []
        for x in range(self.mask.shape[0]):
            for y in range(self.mask.shape[1]):
                if self.mask[x][y] and not blurring_mask[x, y]:
                    frame, kernel_frame = self.frame_at_coords((x, y))
                    self.blurring_frame_array.append(frame)
                    self.blurring_frame_kernel_array.append(kernel_frame)

        self.blurring_frame_lengths = np.asarray(list(map(lambda frame: frame.shape[0], self.blurring_frame_array)),
                                                 dtype='int')

    def frame_at_coords(self, coords):
        """
        Parameters
        ----------
        coords: (int, int)
            The image_grid of mask_index_array on which the frame should be centred
        psf_shape: (int, int)
            The shape of the psf for which this frame will be used
        Returns
        -------
        frame: ndarray
            A subset of mask_index_array of shape psf_shape where elements with image_grid outside of image_frame_indexes have
            value -1
        """
        half_x = int(self.kernel_shape[0] / 2)
        half_y = int(self.kernel_shape[1] / 2)

        frame = np.empty(0, dtype='int')
        kernel_frame = np.empty(0, dtype='float64')

        for i in range(self.kernel_shape[0]):
            for j in range(self.kernel_shape[1]):
                x = coords[0] - half_x + i
                y = coords[1] - half_y + j
                if 0 <= x < self.mask_index_array.shape[0] and 0 <= y < self.mask_index_array.shape[1]:
                    value = self.mask_index_array[x, y]
                    if value >= 0 and not self.mask[x, y]:
                        frame = np.append(frame, value)
                        kernel_frame = np.append(kernel_frame, self.kernel[i, j])

        return frame, kernel_frame


sub_grid_size = 4
# psf_shape = (21, 21)
psf_shape = (41, 41)
sersic = light_profiles.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, intensity=0.1,
                                         effective_radius=0.8, sersic_index=4.0)

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size, psf_shape=psf_shape)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size, psf_shape=psf_shape)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size, psf_shape=psf_shape)


# hst_up = profiling_data.setup_class(analysis_path='HSTup', pixel_scales=0.03, sub_grid_size=sub_grid_size, psf_shape=psf_shape)
# ao = profiling_data.setup_class(analysis_path='AO', pixel_scales=0.01, sub_grid_size=sub_grid_size, psf_shape=psf_shape)

@tools.tick_toc_x1
def lsst_solution():
    lsst_convolver = Convolver(mask=lsst.mask, blurring_mask=lsst.masked_image.blurring_mask,
                               kernel=lsst._image_plane_image.psf.trim_around_centre(psf_shape))


@tools.tick_toc_x1
def euclid_solution():
    euclid_convolver = Convolver(mask=euclid.mask, blurring_mask=euclid.masked_image.blurring_mask,
                                 kernel=euclid._image_plane_image.psf.trim_around_centre(psf_shape))


@tools.tick_toc_x1
def hst_solution():
    hst_convolver = Convolver(mask=hst.mask, blurring_mask=hst.masked_image.blurring_mask,
                              kernel=hst._image_plane_image.psf.trim_around_centre(psf_shape))


@tools.tick_toc_x1
def hst_up_solution():
    hst_up_convolver = Convolver(mask=hst_up.mask, blurring_mask=hst_up.masked_image.blurring_mask,
                                 kernel=hst_up._image_plane_image.psf.trim_around_centre(psf_shape))


@tools.tick_toc_x1
def ao_solution():
    ao_convolver = Convolver(mask=ao.mask, blurring_mask=ao.blurring_mask,
                             kernel=ao._image_plane_image.psf.trim_around_centre(psf_shape))


if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    # hst_up_solution()
    # ao_solution()



import numba
import numpy as np
from analysis import galaxy
from analysis import ray_tracing
from profiling import profiling_data
from profiling import tools

from autolens import exc
from profiles import light_profiles
from profiles import mass_profiles


class RegularizationWeighted(object):
    pixels = None
    regularization_coefficients = None
    pix_signal_scale = None

    def pix_signals_from_images(self, image_to_pix, galaxy_image):
        """Compute the (scaled) signal in each pixel, where the signal is the sum of its masked_image-pixel fluxes. \
        These pix-signals are then used to compute the effective regularization_matrix weight of each pixel.

        The pix signals are scaled in the following ways:

        1) Divided by the number of masked_image-pixels in the pixel, to ensure all pixels have the same \
        'relative' signal (i.e. a pixel with 10 images-pixels doesn't have x2 the signal of one with 5).

        2) Divided by the maximum pix-signal, so that all signals vary between 0 and 1. This ensures that the \
        regularizations weights they're used to compute are defined identically for all masked_image units / SNR's.

        3) Raised to the power of the hyper-parameter *signal_scale*, so the method can control the relative \
        contribution of the different regions of regularization_matrix.
        """

        pix_signals = np.zeros((self.pixels,))
        pix_sizes = np.zeros((self.pixels,))

        for image_index in range(galaxy_image.shape[0]):
            pix_signals[image_to_pix[image_index]] += galaxy_image[image_index]
            pix_sizes[image_to_pix[image_index]] += 1

        pix_signals /= pix_sizes
        pix_signals /= max(pix_signals)

        return pix_signals ** self.pix_signal_scale

    def regularization_weights_from_pix_signals(self, pix_signals):
        """Compute the regularization_matrix weights, which represent the effective regularization_matrix coefficient of every \
        pixel. These are computed using the (scaled) pix-signal in each pixel.

        Two regularization_matrix coefficients are used which mappers to:

        1) pix_signals - This regularizes pix-plane pixels with a high pix-signal (i.e. where the pix is).
        2) 1.0 - pix_signals - This regularizes pix-plane pixels with a low pix-signal (i.e. background sky)
        """
        return (self.regularization_coefficients[0] * pix_signals +
                self.regularization_coefficients[1] * (1.0 - pix_signals)) ** 2.0

    def regularization_matrix_from_pix_neighbors(self, regularization_weights, pix_neighbors):
        """
        Setup a weighted regularization_matrix matrix, where all pixels are regularized with one another in both \
        directions using a different effective regularization_matrix coefficient.

        Parameters
        ----------
        regularization_weights : list(float)
            The regularization_matrix weight of each pixel
        pix_neighbors : [[]]
            A list of the neighbors of each pixel.
        """

        regularization_matrix = np.zeros(shape=(self.pixels, self.pixels))

        reg_weight = regularization_weights ** 2.0

        for i in range(self.pixels):
            for j in pix_neighbors[i]:
                regularization_matrix[i, i] += reg_weight[j]
                regularization_matrix[j, j] += reg_weight[j]
                regularization_matrix[i, j] -= reg_weight[j]
                regularization_matrix[j, i] -= reg_weight[j]

        return regularization_matrix


class Pixelization(object):

    def __init__(self, pixels=100, regularization_coefficients=(1.0,)):
        """
        Abstract base class for a inversion, which discretizes a set of masked_image and sub grid grid into \
        pixels. These pixels incorrect_fit an masked_image using a linear inversion, where a regularization_matrix matrix
        enforces smoothness between pixel values.

        A number of 1D and 2D arrays are used to represent mappings betwen masked_image, sub, pix, and cluster pixels. The \
        nomenclature here follows grid_to_grid, such that it maps the index of a value on one grid to another. For \
        howtolens:

        - pix_to_image[2] = 5 tells us that the 3rd inversion-pixel maps to the 6th masked_image-pixel.
        - sub_to_pixelization[4,2] = 2 tells us that the 5th sub-pixel maps to the 3rd inversion-pixel.

        Parameters
        ----------
        pixels : int
            The number of pixels in the inversion.
        regularization_coefficients : (float,)
            The regularization_matrix coefficients used to smooth the pix reconstructed_image.
        """
        self.pixels = pixels
        self.regularization_coefficients = regularization_coefficients


class RectangularRegWeight(Pixelization, RegularizationWeighted):

    def __init__(self, shape=(3, 3), regularization_coefficients=(1.0, 1.0), pix_signal_scale=1.0):
        """A rectangular inversion where pixels appear on a Cartesian, uniform and rectangular grid \
        of  shape (rows, columns).

        Like an masked_image grid, the indexing of the rectangular grid begins in the top-left corner and goes right and down.

        Parameters
        -----------
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
        regularization_coefficients : (float,)
            The regularization_matrix coefficients used to smooth the pix reconstructed_image.
        """

        if shape[0] <= 2 or shape[1] <= 2:
            raise exc.PixelizationException('The rectangular inversion must be at least dimensions 3x3')

        super(RectangularRegWeight, self).__init__(shape[0] * shape[1], regularization_coefficients)

        self.shape = shape
        self.pix_signal_scale = pix_signal_scale

    class Geometry(object):

        def __init__(self, x_min, x_max, x_pixel_scale, y_min, y_max, y_pixel_scale):
            """The geometry of a rectangular grid, defining where the grids top-left, top-right, bottom-left and \
            bottom-right corners are in arc seconds. The arc-second size of each rectangular pixel is also computed.

            Parameters
            -----------

            """
            self.x_min = x_min
            self.x_max = x_max
            self.x_pixel_scale = x_pixel_scale
            self.y_min = y_min
            self.y_max = y_max
            self.y_pixel_scale = y_pixel_scale

        def arc_second_to_pixel_index_x(self, coordinate):
            return np.floor((coordinate - self.x_min) / self.x_pixel_scale)

        def arc_second_to_pixel_index_y(self, coordinate):
            return np.floor((coordinate - self.y_min) / self.y_pixel_scale)

    def neighbors_from_pixelization(self):
        """Compute the neighbors of every pixel as a list of the pixel index's each pixel shares a vertex with.

        The uniformity of the rectangular grid's geometry is used to compute this.
        """

        def compute_corner_neighbors(pix_neighbors):

            pix_neighbors[0] = [1, self.shape[1]]
            pix_neighbors[self.shape[1] - 1] = [self.shape[1] - 2, self.shape[1] + self.shape[1] - 1]
            pix_neighbors[self.pixels - self.shape[1]] = [self.pixels - self.shape[1] * 2,
                                                          self.pixels - self.shape[1] + 1]
            pix_neighbors[self.pixels - 1] = [self.pixels - self.shape[1] - 1, self.pixels - 2]

            return pix_neighbors

        def compute_top_edge_neighbors(pix_neighbors):

            for pix in range(1, self.shape[1] - 1):
                pixel_index = pix
                pix_neighbors[pixel_index] = [pixel_index - 1, pixel_index + 1, pixel_index + self.shape[1]]

            return pix_neighbors

        def compute_left_edge_neighbors(pix_neighbors):

            for pix in range(1, self.shape[0] - 1):
                pixel_index = pix * self.shape[1]
                pix_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index + 1, pixel_index + self.shape[1]]

            return pix_neighbors

        def compute_right_edge_neighbors(pix_neighbors):

            for pix in range(1, self.shape[0] - 1):
                pixel_index = pix * self.shape[1] + self.shape[1] - 1
                pix_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index - 1, pixel_index + self.shape[1]]

            return pix_neighbors

        def compute_bottom_edge_neighbors(pix_neighbors):

            for pix in range(1, self.shape[1] - 1):
                pixel_index = self.pixels - pix - 1
                pix_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index - 1, pixel_index + 1]

            return pix_neighbors

        def compute_central_neighbors(pix_neighbors):

            for x in range(1, self.shape[0] - 1):
                for y in range(1, self.shape[1] - 1):
                    pixel_index = x * self.shape[1] + y
                    pix_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index - 1, pixel_index + 1,
                                                  pixel_index + self.shape[1]]

            return pix_neighbors

        pixel_neighbors = [[] for _ in range(self.pixels)]

        pixel_neighbors = compute_corner_neighbors(pixel_neighbors)
        pixel_neighbors = compute_top_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_left_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_right_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_bottom_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_central_neighbors(pixel_neighbors)

        return pixel_neighbors

    def geometry_from_pix_sub_grid(self, pix_sub_grid, buffer=1e-8):
        """Determine the geometry of the rectangular grid, by alligning it with the outer-most pix_grid grid \
        plus a small buffer.

        Parameters
        -----------
        pix_sub_grid : [[float, float]]
            The x and y pix grid (or sub-coordinaates) which are to be matched with their pixels.
        buffer : float
            The size the grid-geometry is extended beyond the most exterior grid.
        """
        x_min = np.min(pix_sub_grid[:, 0]) - buffer
        x_max = np.max(pix_sub_grid[:, 0]) + buffer
        y_min = np.min(pix_sub_grid[:, 1]) - buffer
        y_max = np.max(pix_sub_grid[:, 1]) + buffer
        x_pixel_scale = (x_max - x_min) / self.shape[0]
        y_pixel_scale = (y_max - y_min) / self.shape[1]

        return self.Geometry(x_min, x_max, x_pixel_scale, y_min, y_max, y_pixel_scale)

    def grid_to_pix_from_grid(self, grid, geometry):
        """Compute the mappings between a set of masked_image pixels (or sub-pixels) and pixels, using the masked_image's
        traced pix-plane grid (or sub-grid) and the uniform rectangular inversion's geometry.

        Parameters
        ----------
        grid : [[float, float]]
            The x and y pix grid (or sub-coordinates) which are to be matched with their pixels.
        geometry : Geometry
            The rectangular pixel grid's geometry.
        """
        grid_to_pix = np.zeros(grid.shape[0], dtype='int')

        for index, pix_coordinate in enumerate(grid):
            x_pixel = geometry.x_arc_seconds_to_pixels(pix_coordinate[0])
            y_pixel = geometry.y_arc_seconds_to_pixels(pix_coordinate[1])

            grid_to_pix[index] = x_pixel * self.shape[1] + y_pixel

        return grid_to_pix

    def grid_to_pix_from_grid_jitted(self, grid, geometry):
        """Compute the mappings between a set of masked_image pixels (or sub-pixels) and pixels, using the masked_image's
        traced pix-plane grid (or sub-grid) and the uniform rectangular inversion's geometry.

        Parameters
        ----------
        grid : [[float, float]]
            The x and y pix grid (or sub-coordinates) which are to be matched with their pixels.
        geometry : Geometry
            The rectangular pixel grid's geometry.
        """
        return self.grid_to_pix_jit(grid, geometry.x_min, geometry.x_pixel_scale, geometry.y_min,
                                    geometry.y_pixel_scale, self.shape[1]).astype(dtype='int')

    @staticmethod
    @numba.jit(nopython=True)
    def grid_to_pix_jit(grid, x_min, x_pixel_scale, y_min, y_pixel_scale, y_shape):

        grid_to_pix = np.zeros(grid.shape[0])

        for i in range(grid.shape[0]):
            x_pixel = np.floor((grid[i, 0] - x_min) / x_pixel_scale)
            y_pixel = np.floor((grid[i, 1] - y_min) / y_pixel_scale)

            grid_to_pix[i] = x_pixel * y_shape + y_pixel

        return grid_to_pix

    def image_to_pix_from_pix_grids(self, grids, borders):
        """
        Compute the inversion matrices of the rectangular inversion by following these steps:

        1) Setup the rectangular grid geometry, by making its corner appear at the higher / lowest x and y pix sub-
        grid.
        2) Pair masked_image and sub-masked_image pixels to the rectangular grid using their traced grid and its geometry.

        Parameters
        ----------

        """
        relocated_grids = borders.relocated_grids_from_grids(grids)
        geometry = self.geometry_from_pix_sub_grid(relocated_grids.sub)
        image_to_pix = self.grid_to_pix_from_grid_jitted(relocated_grids._image_plane_image, geometry)
        return image_to_pix


sub_grid_size = 4

sie = mass_profiles.EllipticalIsothermal(centre=(0.010, 0.032), einstein_radius=1.47, axis_ratio=0.849, phi=73.6)
shear = mass_profiles.ExternalShear(magnitude=0.0663, phi=160.5)
lens_galaxy = galaxy.Galaxy(mass_profile_0=sie, mass_profile_1=shear)

sersic = light_profiles.EllipticalSersicLightProfile()
source_galaxy = galaxy.Galaxy(light_profile=sersic)

pix = RectangularRegWeight(shape=(50, 50))

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size)
ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=sub_grid_size)

lsst_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                 image_plane_grids=lsst.grids)
euclid_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                   image_plane_grids=euclid.grids)
hst_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                image_plane_grids=hst.grids)
hst_up_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                   image_plane_grids=hst_up.grids)
ao_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy], image_plane_grids=ao.grids)

lsst_image = lsst_tracer.source_plane.galaxy_light_profiles_image_from_planes()
euclid_image = lsst_tracer.source_plane.galaxy_light_profiles_image_from_planes()
hst_image = lsst_tracer.source_plane.galaxy_light_profiles_image_from_planes()
hst_up_image = lsst_tracer.source_plane.galaxy_light_profiles_image_from_planes()
ao_image = lsst_tracer.source_plane.galaxy_light_profiles_image_from_planes()

lsst_image_to_pix = pix.image_to_pix_from_pix_grids(grids=lsst_tracer.source_plane.grids, borders=lsst.borders)
euclid_image_to_pix = pix.image_to_pix_from_pix_grids(grids=euclid_tracer.source_plane.grids, borders=euclid.borders)
hst_image_to_pix = pix.image_to_pix_from_pix_grids(grids=hst_tracer.source_plane.grids, borders=hst.borders)
hst_up_image_to_pix = pix.image_to_pix_from_pix_grids(grids=hst_up_tracer.source_plane.grids, borders=hst_up.borders)
ao_image_to_pix = pix.image_to_pix_from_pix_grids(grids=ao_tracer.source_plane.grids, borders=ao.borders)

lsst_pix_signals = pix.pix_signals_from_images(image_to_pix=lsst_image_to_pix, galaxy_image=lsst_image)
euclid_pix_signals = pix.pix_signals_from_images(image_to_pix=euclid_image_to_pix, galaxy_image=euclid_image)
hst_pix_signals = pix.pix_signals_from_images(image_to_pix=hst_image_to_pix, galaxy_image=hst_image)
hst_up_pix_signals = pix.pix_signals_from_images(image_to_pix=hst_up_image_to_pix, galaxy_image=hst_up_image)
ao_pix_signals = pix.pix_signals_from_images(image_to_pix=ao_image_to_pix, galaxy_image=ao_image)

lsst_regularization_weights = pix.regularization_weights_from_pix_signals(lsst_pix_signals)
euclid_regularization_weights = pix.regularization_weights_from_pix_signals(euclid_pix_signals)
hst_regularization_weights = pix.regularization_weights_from_pix_signals(hst_pix_signals)
hst_up_regularization_weights = pix.regularization_weights_from_pix_signals(hst_up_pix_signals)
ao_regularization_weights = pix.regularization_weights_from_pix_signals(ao_pix_signals)

pix_neighbors = pix.neighbors_from_pixelization()


@tools.tick_toc_x1
def lsst_solution():
    pix.regularization_matrix_from_pix_neighbors(lsst_regularization_weights, pix_neighbors)


@tools.tick_toc_x1
def euclid_solution():
    pix.regularization_matrix_from_pix_neighbors(euclid_regularization_weights, pix_neighbors)


@tools.tick_toc_x1
def hst_solution():
    pix.regularization_matrix_from_pix_neighbors(hst_regularization_weights, pix_neighbors)


@tools.tick_toc_x1
def hst_up_solution():
    pix.regularization_matrix_from_pix_neighbors(hst_up_regularization_weights, pix_neighbors)


@tools.tick_toc_x1
def ao_solution():
    pix.regularization_matrix_from_pix_neighbors(ao_regularization_weights, pix_neighbors)


if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()



import logging

import numpy as np

from autolens.imaging import imaging_util

logging.basicConfig()
logger = logging.getLogger(__name__)


# noinspection PyUnresolvedReferences
class ArrayGeometry(object):
    pixel_scales = None

    @property
    def shape_arc_seconds(self):
        return (float(self.pixel_scales[0] * self.shape[0]), float(self.pixel_scales[1] * self.shape[1]))

    @property
    def central_pixel_coordinates(self):
        return float(self.shape[0] - 1) / 2, float(self.shape[1] - 1) / 2

    def grid_arc_seconds_to_grid_pixels(self, grid_arc_seconds):
        """Convert a grid of (y,x) arc second coordinates to a grid of (y,x) pixel values. Pixel coordinates are
        returned as floats such that they include the decimal offset from each pixel's top-left corner.

        The pixel coordinate origin is at the top left corner of the grid, such that the pixel [0,0] corresponds to \
        higher y arc-second coordinate value and lowest x arc-second coordinate.

        The arc-second coordinate origin is at the centre and will correspond to the dimensions of each image divided \
        by two.

        Parameters
        ----------
        grid_arc_seconds: ndarray
            The grid of (y,x) coordinates in arc seconds.
        """
        return imaging_util.grid_arc_seconds_1d_to_grid_pixels_1d(grid_arc_seconds=grid_arc_seconds, shape=self.shape,
                                                                  pixel_scales=self.pixel_scales)

    def grid_arc_seconds_to_grid_pixel_centres(self, grid_arc_seconds):
        """Convert a grid of (y,x) arc second coordinates to a grid of (y,x) pixel values. Pixel coordinates are \
        returned as integers such that they map directly to the pixel they are contained within.

        The pixel coordinate origin is at the top left corner of the grid, such that the pixel [0,0] corresponds to \
        higher y arc-second coordinate value and lowest x arc-second coordinate.

        The arc-second coordinate origin is at the centre and will correspond to the dimensions of each image divided \
        by two.

        Parameters
        ----------
        grid_arc_seconds: ndarray
            The grid of (y,x) coordinates in arc seconds.
        """
        return imaging_util.grid_arc_seconds_1d_to_grid_pixel_centres_1d(grid_arc_seconds=grid_arc_seconds,
                                                                         shape=self.shape,
                                                                         pixel_scales=self.pixel_scales)

    def grid_pixels_to_grid_arc_seconds(self, grid_pixels):
        """Convert a grid of (y,x) arc second coordinates to a grid of (y,x) pixel values. Pixel coordinates are
        returned as floats such that they include the decimal offset from each pixel's top-left corner.

        The pixel coordinate origin is at the top left corner of the grid, such that the pixel [0,0] corresponds to \
        higher y arc-second coordinate value and lowest x arc-second coordinate.

        The arc-second coordinate origin is at the centre and will correspond to the dimensions of each image divided \
        by two.

        Parameters
        ----------
        grid_arc_seconds: ndarray
            The grid of (y,x) coordinates in arc seconds.
        """
        return imaging_util.grid_pixels_1d_to_grid_arc_seconds_1d(grid_pixels=grid_pixels, shape=self.shape,
                                                                  pixel_scales=self.pixel_scales)

    @property
    def grid_1d(self):
        """ The arc second-grid of (y,x) coordinates of every pixel.

        This is defined from the top-left corner, such that the first pixel at location [0, 0] will have a negative x \
        value y value in arc seconds.
        """
        return imaging_util.image_grid_1d_from_shape_and_pixel_scales(shape=self.shape, pixel_scales=self.pixel_scales)

    @property
    def grid_2d(self):
        """ The arc second-grid of (y,x) coordinates of every pixel.

        This is defined from the top-left corner, such that the first pixel at location [0, 0] will have a negative x \
        value y value in arc seconds.
        """
        return imaging_util.image_grid_2d_from_shape_and_pixel_scales(shape=self.shape, pixel_scales=self.pixel_scales)

    @property
    def arc_second_maxima(self):
        return (self.shape_arc_seconds[0]) / 2.0, (self.shape_arc_seconds[1]) / 2.0

    @property
    def arc_second_minima(self):
        return -(self.shape_arc_seconds[0]) / 2.0, -(self.shape_arc_seconds[1]) / 2.0

    @property
    def yticks(self):
        """Compute the yticks labels of this grid, used for plotting the y-axis ticks when visualizing an _data-grid"""
        return np.linspace(-self.shape_arc_seconds[0] / 2.0, self.shape_arc_seconds[0] / 2.0, 4)

    @property
    def xticks(self):
        """Compute the xticks labels of this grid, used for plotting the x-axis ticks when visualizing an _data-grid"""
        return np.linspace(-self.shape_arc_seconds[1] / 2.0, self.shape_arc_seconds[1] / 2.0, 4)


class Array(np.ndarray):

    def __new__(cls, array, *args, **kwargs):
        return np.array(array, dtype='float64').view(cls)

    def __reduce__(self):
        # Get the parent's __reduce__ tuple
        pickled_state = super(Array, self).__reduce__()
        # Create our own tuple to pass to __setstate__
        class_dict = {}
        for key, value in self.__dict__.items():
            class_dict[key] = value
        new_state = pickled_state[2] + (class_dict,)
        # Return a tuple that replaces the parent's __setstate__ tuple with our own
        return pickled_state[0], pickled_state[1], new_state

    # noinspection PyMethodOverriding
    def __setstate__(self, state):

        for key, value in state[-1].items():
            setattr(self, key, value)
        super(Array, self).__setstate__(state[0:-1])

    def resize_around_centre(self, new_shape):

        if new_shape[0] < self.shape[0] and new_shape[1] < self.shape[1]:
            return self.trim_around_centre(new_shape=new_shape)
        elif new_shape[0] > self.shape[0] and new_shape[1] > self.shape[1]:
            return self.pad_around_centre(new_shape=new_shape)
        else:
            raise ValueError('resize_around_centre - one dimensions of the new_shape pads the image, whereas the other'
                             'trims it - make the trimmed dimension larger.')

    def pad_around_centre(self, new_shape):
        """Pad the array to a new shape.

        Parameters
        -----------
        new_shape : (int, int)
            The new two-dimensional shape of the array.
        """
        return self.new_with_array(imaging_util.pad_array_2d_around_centre(self, new_shape))

    def trim_around_centre(self, new_shape):
        """Trim the array to a new shape.

        Parameters
        -----------
        new_shape : (int, int)
            The new two-dimensional shape of the array.
        """
        return self.new_with_array(imaging_util.trim_array_2d_around_centre(self, new_shape))

    def trim_around_region(self, x0, x1, y0, y1):
        """Trim the array to a new shape.
        """
        return self.new_with_array(imaging_util.trim_array_2d_around_region(self, x0, x1, y0, y1))

    def new_with_array(self, array):
        """
        Parameters
        ----------
        array: ndarray
            An ndarray

        Returns
        -------
        new_array: ScaledSquarePixelArray
            A new instance of this class that shares all of this instances attributes with a new ndarray.
        """
        arguments = vars(self)
        arguments.update({"array": array})

        return self.__class__(**arguments)

    @classmethod
    def from_fits(cls, file_path, hdu):
        """
        Loads the data from a .fits file.

        Parameters
        ----------
        file_path : str
            The full path of the fits file.
        hdu : int
            The HDU number in the fits file containing the _data data.
        """
        return cls(imaging_util.numpy_array_from_fits(file_path, hdu))


class ScaledArray(Array, ArrayGeometry):

    # noinspection PyUnusedLocal
    def __init__(self, array):
        """
        Parameters
        ----------
        array: ndarray
            An array representing data (e.g. an _data, noise-mappers, etc.)
        """
        # noinspection PyArgumentList
        super(ScaledArray, self).__init__()

    def map(self, func):
        for y in range(self.shape[0]):
            for x in range(self.shape[1]):
                func(y, x)


class ScaledSquarePixelArray(ScaledArray):
    """
    Class storing the grids for 2D pixel grids (e.g. _data, PSF, signal_to_noise_ratio).
    """

    # noinspection PyUnusedLocal
    def __init__(self, array, pixel_scale):
        """
        Parameters
        ----------
        array: ndarray
            An array representing data (e.g. an _data, noise-mappers, etc.)
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        """
        # noinspection PyArgumentList
        self.pixel_scale = pixel_scale
        super(ScaledSquarePixelArray, self).__init__(array=array)

    def __array_finalize__(self, obj):
        if hasattr(obj, "pixel_scale"):
            self.pixel_scale = obj.pixel_scale

    @property
    def pixel_scales(self):
        return self.pixel_scale, self.pixel_scale

    @classmethod
    def from_fits_with_pixel_scale(cls, file_path, hdu, pixel_scale):
        """
        Loads the data from a .fits file.

        Parameters
        ----------
        file_path : str
            The full path of the fits file.
        hdu : int
            The HDU number in the fits file containing the _data data.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        """
        return cls(imaging_util.numpy_array_from_fits(file_path, hdu), pixel_scale)

    @classmethod
    def single_value(cls, value, shape, pixel_scale):
        """
        Creates an instance of Array and fills it with a single value

        Parameters
        ----------
        value: float
            The value with which the array should be filled
        shape: (int, int)
            The shape of the array
        pixel_scale: float
            The scale of a pixel in arc seconds

        Returns
        -------
        array: ScaledSquarePixelArray
            An array filled with a single value
        """
        array = np.ones(shape) * value
        return cls(array, pixel_scale)

    def flatten(self, order='C'):
        """
        Returns
        -------
        flat_scaled_array: ScaledSquarePixelArray
            A copy of this array flattened to 1D
        """
        return self.new_with_array(super(ScaledSquarePixelArray, self).flatten(order))

    def __eq__(self, other):
        super_result = super(ScaledSquarePixelArray, self).__eq__(other)
        try:
            return super_result.all()
        except AttributeError:
            return super_result


class ScaledRectangularPixelArray(ScaledArray):
    """
    Class storing the grids for 2D pixel grids (e.g. _data, PSF, signal_to_noise_ratio).
    """

    # noinspection PyUnusedLocal
    def __init__(self, array, pixel_scales):
        """
        Parameters
        ----------
        array: ndarray
            An array representing data (e.g. an _data, noise-mappers, etc.)
        pixel_scales : (float, float)
            The arc-second to pixel conversion factor of each pixel.
        """
        self.pixel_scales = pixel_scales
        # noinspection PyArgumentList
        super(ScaledRectangularPixelArray, self).__init__(array=array)

    def __array_finalize__(self, obj):
        if hasattr(obj, "pixel_scales"):
            self.pixel_scales = obj.pixel_scales

    @classmethod
    def single_value(cls, value, shape, pixel_scales):
        """
        Creates an instance of Array and fills it with a single value

        Parameters
        ----------
        value: float
            The value with which the array should be filled
        shape: (int, int)
            The shape of the array
        pixel_scales: (float, float)
            The scale of a pixel in arc seconds

        Returns
        -------
        array: ScaledSquarePixelArray
            An array filled with a single value
        """
        array = np.ones(shape) * value
        return cls(array, pixel_scales)

    @classmethod
    def from_fits_with_pixel_scale(cls, file_path, hdu, pixel_scales):
        """
        Loads the data from a .fits file.

        Parameters
        ----------
        file_path : str
            The full path of the fits file.
        hdu : int
            The HDU number in the fits file containing the _data data.
        pixel_scales: (float, float)
            The arc-second to pixel conversion factor of each pixel.
        """
        return cls(imaging_util.numpy_array_from_fits(file_path, hdu), pixel_scales)

    def flatten(self, order='C'):
        """
        Returns
        -------
        flat_scaled_array: ScaledSquarePixelArray
            A copy of this array flattened to 1D
        """
        return self.new_with_array(super().flatten(order))

    def __eq__(self, other):
        super_result = super(ScaledRectangularPixelArray, self).__eq__(other)
        try:
            return super_result.all()
        except AttributeError:
            return super_result



import numpy as np
from profiling import profiling_data
from profiling import tools

from profiles import geometry_profiles


class SphericalPowerLaw(geometry_profiles.SphericalProfile):

    def __init__(self, centre=(0.0, 0.0), einstein_radius=1.0, slope=2.0):
        """
        Represents a spherical power-law density distribution.

        Parameters
        ----------
        centre: (float, float)
            The (x,y) coordinates of the centre of the profile.
        einstein_radius : float
            Einstein radius of power-law mass profiles
        slope : float
            power-law density slope of mass profiles
        """
        super(SphericalPowerLaw, self).__init__(centre)
        self.einstein_radius = einstein_radius
        self.slope = slope

    @property
    def einstein_radius_rescaled(self):
        """Rescale the einstein radius by slope and axis_ratio, to reduce its degeneracy with other mass-profiles
        parameters"""
        return ((3 - self.slope) / (1 + self.axis_ratio)) * self.einstein_radius ** (self.slope - 1)

    def deflections_from_grid(self, grid):
        eta = self.grid_to_radius(grid)
        deflection_r = 2.0 * self.einstein_radius_rescaled * \
                       np.divide(np.power(eta, (3.0 - self.slope)), np.multiply((3.0 - self.slope), eta))
        return deflection_r


power_law = SphericalPowerLaw(centre=(0.0, 0.0), einstein_radius=1.4, slope=2.0)

subgrd_size = 4

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=subgrd_size)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=subgrd_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=subgrd_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=subgrd_size)
ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=subgrd_size)


@tools.tick_toc_x10
def lsst_solution():
    power_law.deflections_from_grid(grid=lsst.coords.sub_grid_coords)


@tools.tick_toc_x10
def euclid_solution():
    power_law.deflections_from_grid(grid=euclid.coords.sub_grid_coords)


@tools.tick_toc_x10
def hst_solution():
    power_law.deflections_from_grid(grid=hst.coords.sub_grid_coords)


@tools.tick_toc_x10
def hst_up_solution():
    power_law.deflections_from_grid(grid=hst_up.coords.sub_grid_coords)


@tools.tick_toc_x10
def ao_solution():
    power_law.deflections_from_grid(grid=ao.coords.sub_grid_coords)


if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()



import numpy as np
import scipy.spatial
import sklearn.cluster

from autolens import exc
from autolens.imaging import scaled_array
from autolens.imaging import mask
from autolens.inversion import mappers


class Pixelization(object):

    def __init__(self, pixels=100):
        """ Abstract base class for a pixelization, which discretizes a set of coordinates (e.g. an _data-grid) into \
        pixels.

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        """
        self.pixels = pixels

    def mapper_from_grids_and_borders(self, grids, borders):
        raise NotImplementedError("pixelization_mapper_from_grids_and_borders should be overridden")


class Rectangular(Pixelization):

    def __init__(self, shape=(3, 3)):
        """A rectangular pixelization, where pixels are defined on a Cartesian and uniform grid of shape (rows, columns).

        Like an _data, the indexing of the rectangular grid begins in the top-left corner and goes right and down.

        Parameters
        -----------
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
        """

        if shape[0] <= 2 or shape[1] <= 2:
            raise exc.PixelizationException('The rectangular pixelization must be at least dimensions 3x3')

        self.shape = (int(shape[0]), int(shape[1]))

        super(Rectangular, self).__init__(self.shape[0] * self.shape[1])

    class Geometry(scaled_array.ArrayGeometry):

        def __init__(self, shape, pixel_scales):
            """The geometry of a rectangular grid

            Parameters
            -----------
            y_min : float
                Minimum arcsecond y value of the pixelization (e.g. the top-edge).
            y_max : float
                Maximum arcsecond y value of the pixelization (e.g. the bottom-edge).
            x_min : float
                Minimum arcsecond x value of the pixelization (e.g. the left-edge).
            x_max : float
                Maximum arcsecond x value of the pixelization (e.g. the right-edge).
            pixel_scales : (float, float)
                The pixel-to-arcsecond scale of a pixel in the y and x directions.
            """
            self.shape = shape
            self.pixel_scales = pixel_scales

        @property
        def pixel_centres(self):
            return self.grid_1d

    def geometry_from_grid(self, grid, buffer=1e-8):
        """Determine the geometry of the rectangular grid, by alligning it with the outer-most pixels on a grid \
        plus a small buffer.

        Parameters
        -----------
        grid : [[float, float]]
            The x and y pix grid (or sub-coordinates) which are to be matched with their pixels.
        buffer : float
            The size the grid-geometry is extended beyond the most exterior grid.
        """
        y_min = np.min(grid[:, 0]) - buffer
        y_max = np.max(grid[:, 0]) + buffer
        x_min = np.min(grid[:, 1]) - buffer
        x_max = np.max(grid[:, 1]) + buffer
        pixel_scales = ((y_max - y_min) / self.shape[0], (x_max - x_min) / self.shape[1])

        return self.Geometry(shape=self.shape, pixel_scales=pixel_scales)

    def neighbors_from_pixelization(self):
        """Compute the neighbors of every pixel as a list of the pixel index's each pixel shares a vertex with.

        The uniformity of the rectangular grid's geometry is used to compute this.
        """

        def compute_corner_neighbors(pixel_neighbors):

            pixel_neighbors[0] = [1, self.shape[1]]
            pixel_neighbors[self.shape[1] - 1] = [self.shape[1] - 2, self.shape[1] + self.shape[1] - 1]
            pixel_neighbors[self.pixels - self.shape[1]] = [self.pixels - self.shape[1] * 2,
                                                            self.pixels - self.shape[1] + 1]
            pixel_neighbors[self.pixels - 1] = [self.pixels - self.shape[1] - 1, self.pixels - 2]

            return pixel_neighbors

        def compute_top_edge_neighbors(pixel_neighbors):

            for pix in range(1, self.shape[1] - 1):
                pixel_index = pix
                pixel_neighbors[pixel_index] = [pixel_index - 1, pixel_index + 1, pixel_index + self.shape[1]]

            return pixel_neighbors

        def compute_left_edge_neighbors(pixel_neighbors):

            for pix in range(1, self.shape[0] - 1):
                pixel_index = pix * self.shape[1]
                pixel_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index + 1,
                                                pixel_index + self.shape[1]]

            return pixel_neighbors

        def compute_right_edge_neighbors(pixel_neighbors):

            for pix in range(1, self.shape[0] - 1):
                pixel_index = pix * self.shape[1] + self.shape[1] - 1
                pixel_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index - 1,
                                                pixel_index + self.shape[1]]

            return pixel_neighbors

        def compute_bottom_edge_neighbors(pixel_neighbors):

            for pix in range(1, self.shape[1] - 1):
                pixel_index = self.pixels - pix - 1
                pixel_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index - 1, pixel_index + 1]

            return pixel_neighbors

        def compute_central_neighbors(pixel_neighbors):

            for x in range(1, self.shape[0] - 1):
                for y in range(1, self.shape[1] - 1):
                    pixel_index = x * self.shape[1] + y
                    pixel_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index - 1, pixel_index + 1,
                                                    pixel_index + self.shape[1]]

            return pixel_neighbors

        pixel_neighbors = [[] for _ in range(self.pixels)]

        pixel_neighbors = compute_corner_neighbors(pixel_neighbors)
        pixel_neighbors = compute_top_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_left_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_right_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_bottom_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_central_neighbors(pixel_neighbors)

        return pixel_neighbors

    def mapper_from_grids(self, grids):
        """Setup the pixelization mapper of this rectangular pixelization as follows:

        This first relocateds all grid-coordinates, such that any which tracer beyond its border (e.g. due to high \
        levels of demagnification) are relocated to the border.

        Parameters
        ----------
        grids: mask.ImagingGrids
            A collection of grid describing the observed _data's pixel coordinates (includes an _data and sub grid).
        borders : mask.ImagingGridBorders
            The borders of the grids (defined by their _data-plane mask).
        """
        geometry = self.geometry_from_grid(grids.sub)
        pixel_neighbors = self.neighbors_from_pixelization()
        return mappers.RectangularMapper(pixels=self.pixels, grids=grids, pixel_neighbors=pixel_neighbors,
                                         shape=self.shape, geometry=geometry)


    def mapper_from_grids_and_borders(self, grids, borders):
        """Setup the pixelization mapper of this rectangular pixelization as follows:

        This first relocateds all grid-coordinates, such that any which tracer beyond its border (e.g. due to high \
        levels of demagnification) are relocated to the border.

        Parameters
        ----------
        grids: mask.ImagingGrids
            A collection of grid describing the observed _data's pixel coordinates (includes an _data and sub grid).
        borders : mask.ImagingGridBorders
            The borders of the grids (defined by their _data-plane mask).
        """
        try:
            relocated_grids = borders.relocated_grids_from_grids(grids)
        except ValueError:
            relocated_grids = grids

        geometry = self.geometry_from_grid(relocated_grids.sub)
        pixel_neighbors = self.neighbors_from_pixelization()
        return mappers.RectangularMapper(pixels=self.pixels, grids=relocated_grids, pixel_neighbors=pixel_neighbors,
                                         shape=self.shape, geometry=geometry)


class Voronoi(Pixelization):

    def __init__(self, pixels=100):
        """Abstract base class for a Voronoi pixelization, which represents pixels as an irregular grid of Voronoi \
         pixels which can form any shape, size or tesselation.

         The traced _data-pixels are paired to Voronoi pixels as the nearest-neighbors of the Voronoi pixel-centers.

         Parameters
         ----------
         pixels : int
             The number of pixels in the pixelization.
         """
        super(Voronoi, self).__init__(pixels)

    @staticmethod
    def voronoi_from_pixel_centers(pixel_centers):
        """Compute the Voronoi grid of the pixelization, using the pixel centers.

        Parameters
        ----------
        pixel_centers : ndarray
            The x and y image_grid to derive the Voronoi grid_coords.
        """
        return scipy.spatial.Voronoi(pixel_centers, qhull_options='Qbb Qc Qx Qm')

    def neighbors_from_pixelization(self, ridge_points):
        """Compute the neighbors of every pixel as a list of the pixel index's each pixel shares a vertex with.

        The ridge points of the Voronoi grid are used to derive this.

        Parameters
        ----------
        ridge_points : scipy.spatial.Voronoi.ridge_points
            Each Voronoi-ridge (two indexes representing a pixel mapping_matrix).
        """
        pixel_neighbors = [[] for _ in range(self.pixels)]

        for pair in reversed(ridge_points):
            pixel_neighbors[pair[0]].append(pair[1])
            pixel_neighbors[pair[1]].append(pair[0])

        return pixel_neighbors


class Cluster(Voronoi):

    def __init__(self, pixels):
        """A cluster pixelization, which represents pixels as a Voronoi grid (see Voronoi). For this pixelization, the \
        pixel-centers are derived from a sparse-grid in the observed _data which are lensed to form the pixel centers.

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        """
        super(Cluster, self).__init__(pixels)

    def mapper_from_grids(self, grids, pixel_centers, image_to_voronoi):
        """Setup the pixelization mapper of the cluster pixelization.

        This first relocateds all grid-coordinates, such that any which tracer beyond its border (e.g. due to high \
        levels of demagnification) are relocated to the border.

        Parameters
        ----------
        grids: mask.ImagingGrids
            A collection of grid describing the observed _data's pixel coordinates (includes an _data and sub grid).
        borders : mask.ImagingGridBorders
            The borders of the grids (defined by their _data-plane mask).
        pixel_centers : ndarray
            The center of each Voronoi pixel, computed from an traced _data-plane grid.
        image_to_voronoi : ndarray
            The mapping of each _data pixel to Voronoi pixels.
        """

        voronoi_to_pixelization = np.arange(0, self.pixels)
        voronoi = self.voronoi_from_pixel_centers(pixel_centers)
        pixel_neighbors = self.neighbors_from_pixelization(voronoi.ridge_points)

        return mappers.VoronoiMapper(pixels=self.pixels, grids=grids,
                                     pixel_neighbors=pixel_neighbors,
                                     pixel_centers=pixel_centers, voronoi=voronoi,
                                     voronoi_to_pixelization=voronoi_to_pixelization,
                                     image_to_voronoi=image_to_voronoi)

    def mapper_from_grids_and_borders(self, grids, borders, pixel_centers, image_to_voronoi):
        """Setup the pixelization mapper of the cluster pixelization.

        This first relocateds all grid-coordinates, such that any which tracer beyond its border (e.g. due to high \
        levels of demagnification) are relocated to the border.

        Parameters
        ----------
        grids: mask.ImagingGrids
            A collection of grid describing the observed _data's pixel coordinates (includes an _data and sub grid).
        borders : mask.ImagingGridBorders
            The borders of the grids (defined by their _data-plane mask).
        pixel_centers : ndarray
            The center of each Voronoi pixel, computed from an traced _data-plane grid.
        image_to_voronoi : ndarray
            The mapping of each _data pixel to Voronoi pixels.
        """

        relocated_grids = borders.relocated_grids_from_grids(grids)
        voronoi_to_pixelization = np.arange(0, self.pixels)
        voronoi = self.voronoi_from_pixel_centers(pixel_centers)
        pixel_neighbors = self.neighbors_from_pixelization(voronoi.ridge_points)

        return mappers.VoronoiMapper(pixels=self.pixels, grids=relocated_grids,
                                     pixel_neighbors=pixel_neighbors,
                                     pixel_centers=pixel_centers, voronoi=voronoi,
                                     voronoi_to_pixelization=voronoi_to_pixelization,
                                     image_to_voronoi=image_to_voronoi)


class Amorphous(Voronoi):

    def __init__(self, pixels):
        """
        An amorphous pixelization, which represents pixels as a set of centers where all of the \
        nearest-neighbor pix-grid (i.e. traced masked_image-pixels) are mapped to them.

        For this pixelization, a set of cluster-pixels (defined in the masked_image-plane as a cluster uniform grid of \
        masked_image-pixels) are used to determine a set of pix-plane grid. These grid are then fed into a \
        weighted k-means clustering algorithm, such that the pixel centers adapt to the unlensed pix \
        surface-brightness profile.

        Parameters
        ----------
        pixels : int
            The number of pixels in the pixelization.
        regularization_coefficients : (float,)
            The regularization_matrix coefficients used to smooth the pix reconstructed_image.
        """
        super(Amorphous, self).__init__(pixels)

    def kmeans_cluster(self, cluster_grid):
        """Perform k-means clustering on the cluster_grid to compute the k-means clusters which represent \
        pixels.

        Parameters
        ----------
        cluster_grid : ndarray
            The x and y cluster-grid which are used to derive the k-means pixelization.
        """
        kmeans = sklearn.cluster.KMeans(self.pixels)
        km = kmeans.fit(cluster_grid)
        return km.cluster_centers_, km.labels_


class PixelizationGrid(object):

    def __init__(self, shape=(3, 3)):
        self.shape = shape

    def coordinate_grid_within_annulus(self, inner_radius, outer_radius, centre=(0., 0.)):

        y_pixel_scale = 2.0 * outer_radius / self.shape[0]
        x_pixel_scale = 2.0 * outer_radius / self.shape[1]

        central_pixel = float(self.shape[0] - 1) / 2, float(self.shape[1] - 1) / 2

        pixel_count = 0

        for y in range(self.shape[0]):
            for x in range(self.shape[1]):

                y_arcsec = ((y - central_pixel[1]) * y_pixel_scale) - centre[0]
                x_arcsec = ((x - central_pixel[0]) * x_pixel_scale) - centre[1]
                radius_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

                if radius_arcsec < outer_radius or radius_arcsec > inner_radius:
                    pixel_count += 1

        coordinates_array = np.zeros((pixel_count, 2))

        pixel_count = 0

        for y in range(self.shape[0]):
            for x in range(self.shape[1]):

                y_arcsec = -((y - central_pixel[0]) * y_pixel_scale) - centre[0]
                x_arcsec = ((x - central_pixel[1]) * x_pixel_scale) - centre[1]
                radius_arcsec = np.sqrt(x_arcsec ** 2 + y_arcsec ** 2)

                if radius_arcsec < outer_radius or radius_arcsec > inner_radius:
                    coordinates_array[pixel_count, 0] = y_arcsec
                    coordinates_array[pixel_count, 1] = x_arcsec

                    pixel_count += 1

        return coordinates_array



class MockTracer(object):

    def __init__(self, image, blurring_image, has_light_profile, has_pixelization, has_hyper_galaxy,
                 has_grid_mappers=False):
        self.image = image
        self.blurring_image = blurring_image
        self.has_light_profile = has_light_profile
        self.has_pixelization = has_pixelization
        self.has_hyper_galaxy = has_hyper_galaxy
        self.has_grid_mappers = has_grid_mappers

    @property
    def all_planes(self):
        return []

    @property
    def _image_plane_image(self):
        return self.image

    @property
    def _image_plane_blurring_image(self):
        return self.blurring_image

    @property
    def mappers_of_planes(self):
        return [MockMapper()]

    @property
    def regularization_of_planes(self):
        return [MockMapper()]

    @property
    def hyper_galaxies(self):
        return [MockHyperGalaxy(), MockHyperGalaxy()]


import numpy as np
import pytest

from autolens.inversion import pixelizations


class MockGeometry(object):

    def __init__(self, y_min, y_max, x_min, x_max, y_pixel_scale, x_pixel_scale):
        self.y_min = y_min
        self.y_max = y_max
        self.x_min = x_min
        self.x_max = x_max
        self.y_pixel_scale = y_pixel_scale
        self.x_pixel_scale = x_pixel_scale


@pytest.fixture(name="five_pixels")
def make_five_pixels():
    return np.array([[0, 0], [0, 1], [1, 0], [1, 1], [1, 2]])


@pytest.fixture(name="three_pixels")
def make_three_pixels():
    return np.array([[0, 0], [0, 1], [1, 0]])


class TestRectangular:

    class TestConstructor:

        def test__number_of_pixels_and_regularization_set_up_correctly(self):
            pix = pixelizations.Rectangular(shape=(3, 3))

            assert pix.shape == (3, 3)
            assert pix.pixels == 9

    class TestGeometry:

        def test__3x3_grid__buffer_is_small__grid_give_min_minus_1_max_1__sets_up_geometry_correctly(self):

            pix = pixelizations.Rectangular(shape=(3, 3))

            pixelization_grid = np.array([[1.0, -1.0], [1.0, 0.0], [1.0, 1.0],
                                          [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                          [-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0]])

            geometry = pix.geometry_from_grid(pixelization_grid, buffer=1e-8)

            assert geometry.shape == (3,3)
            assert geometry.pixel_scales == pytest.approx((2./3., 2./3.), 1e-2)

        def test__3x3_grid__same_as_above_change_buffer(self):
            pix = pixelizations.Rectangular(shape=(3, 3))

            pixelization_grid = np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                          [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                          [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            geometry = pix.geometry_from_grid(pixelization_grid, buffer=1e-4)

            assert geometry.shape == (3,3)
            assert geometry.pixel_scales == pytest.approx((2./3., 2./3.), 1e-2)

        def test__5x4_grid__buffer_is_small(self):

            pix = pixelizations.Rectangular(shape=(5, 4))

            pixelization_grid = np.array([[1.0, -1.0], [1.0, 0.0], [1.0, 1.0],
                                          [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                          [-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0]])

            geometry = pix.geometry_from_grid(pixelization_grid, buffer=1e-8)

            assert geometry.shape == (5,4)
            assert geometry.pixel_scales == pytest.approx((2./5., 2./4.), 1e-2)

        def test__3x3_grid__larger_range_of_grid(self):

            pix = pixelizations.Rectangular(shape=(3, 3))

            pixelization_grid = np.array([[2.0, 1.0], [4.0, 3.0], [6.0, 5.0], [8.0, 7.0]])

            geometry = pix.geometry_from_grid(pixelization_grid, buffer=1e-8)

            assert geometry.shape == (3,3)
            assert geometry.pixel_scales == pytest.approx((6./3., 6./3.), 1e-2)

    class TestPixelCentres:

        def test__3x3_grid__pixel_centres(self):

            pix = pixelizations.Rectangular(shape=(3, 3))

            pixelization_grid = np.array([[1.0, -1.0], [1.0, 0.0], [1.0, 1.0],
                                          [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                          [-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0]])

            geometry = pix.geometry_from_grid(pixelization_grid, buffer=1e-8)

            assert geometry.pixel_centres == pytest.approx(np.array([[2./3., -2./3.], [2./3., 0.0], [2./3., 2./3.],
                                                                     [ 0.0, -2./3.], [ 0.0, 0.0], [ 0.0, 2./3.],
                                                                     [-2./3., -2./3.], [-2./3., 0.0], [-2./3., 2./3.]]))

        def test__4x3_grid__pixel_centres(self):

            pix = pixelizations.Rectangular(shape=(4, 3))

            pixelization_grid = np.array([[1.0, -1.0], [1.0, 0.0], [1.0, 1.0],
                                          [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                          [-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0]])

            geometry = pix.geometry_from_grid(pixelization_grid, buffer=1e-8)

            assert geometry.pixel_centres == pytest.approx(np.array([[0.75, -2./3.], [0.75, 0.0], [0.75, 2./3.],
                                                                     [0.25, -2./3.], [0.25, 0.0], [0.25, 2./3.],
                                                                     [-0.25, -2./3.], [-0.25, 0.0], [-0.25, 2./3.],
                                                                     [-0.75, -2./3.], [-0.75, 0.0],[-0.75, 2./ 3.],]))

    class TestPixelNeighbors:

        def test__compute_pixel_neighbors__3x3_grid(self):
            # |0|1|2|
            # |3|4|5|
            # |6|7|8|

            pix = pixelizations.Rectangular(shape=(3, 3))

            pixel_neighbors = pix.neighbors_from_pixelization()

            assert pixel_neighbors[0] == [1, 3]
            assert pixel_neighbors[1] == [0, 2, 4]
            assert pixel_neighbors[2] == [1, 5]
            assert pixel_neighbors[3] == [0, 4, 6]
            assert pixel_neighbors[4] == [1, 3, 5, 7]
            assert pixel_neighbors[5] == [2, 4, 8]
            assert pixel_neighbors[6] == [3, 7]
            assert pixel_neighbors[7] == [4, 6, 8]
            assert pixel_neighbors[8] == [5, 7]

        def test__compute_pixel_neighbors__3x4_grid(self):
            # |0|1| 2| 3|
            # |4|5| 6| 7|
            # |8|9|10|11|

            pix = pixelizations.Rectangular(shape=(3, 4))

            pixel_neighbors = pix.neighbors_from_pixelization()

            assert pixel_neighbors[0] == [1, 4]
            assert pixel_neighbors[1] == [0, 2, 5]
            assert pixel_neighbors[2] == [1, 3, 6]
            assert pixel_neighbors[3] == [2, 7]
            assert pixel_neighbors[4] == [0, 5, 8]
            assert pixel_neighbors[5] == [1, 4, 6, 9]
            assert pixel_neighbors[6] == [2, 5, 7, 10]
            assert pixel_neighbors[7] == [3, 6, 11]
            assert pixel_neighbors[8] == [4, 9]
            assert pixel_neighbors[9] == [5, 8, 10]
            assert pixel_neighbors[10] == [6, 9, 11]
            assert pixel_neighbors[11] == [7, 10]

        def test__compute_pixel_neighbors__4x3_grid(self):
            # |0| 1| 2|
            # |3| 4| 5|
            # |6| 7| 8|
            # |9|10|11|

            pix = pixelizations.Rectangular(shape=(4, 3))

            pixel_neighbors = pix.neighbors_from_pixelization()

            assert pixel_neighbors[0] == [1, 3]
            assert pixel_neighbors[1] == [0, 2, 4]
            assert pixel_neighbors[2] == [1, 5]
            assert pixel_neighbors[3] == [0, 4, 6]
            assert pixel_neighbors[4] == [1, 3, 5, 7]
            assert pixel_neighbors[5] == [2, 4, 8]
            assert pixel_neighbors[6] == [3, 7, 9]
            assert pixel_neighbors[7] == [4, 6, 8, 10]
            assert pixel_neighbors[8] == [5, 7, 11]
            assert pixel_neighbors[9] == [6, 10]
            assert pixel_neighbors[10] == [7, 9, 11]
            assert pixel_neighbors[11] == [8, 10]

        def test__compute_pixel_neighbors__4x4_grid(self):
            # |0 | 1| 2| 3|
            # |4 | 5| 6| 7|
            # |8 | 9|10|11|
            # |12|13|14|15|

            pix = pixelizations.Rectangular(shape=(4, 4))

            pixel_neighbors = pix.neighbors_from_pixelization()

            assert pixel_neighbors[0] == [1, 4]
            assert pixel_neighbors[1] == [0, 2, 5]
            assert pixel_neighbors[2] == [1, 3, 6]
            assert pixel_neighbors[3] == [2, 7]
            assert pixel_neighbors[4] == [0, 5, 8]
            assert pixel_neighbors[5] == [1, 4, 6, 9]
            assert pixel_neighbors[6] == [2, 5, 7, 10]
            assert pixel_neighbors[7] == [3, 6, 11]
            assert pixel_neighbors[8] == [4, 9, 12]
            assert pixel_neighbors[9] == [5, 8, 10, 13]
            assert pixel_neighbors[10] == [6, 9, 11, 14]
            assert pixel_neighbors[11] == [7, 10, 15]
            assert pixel_neighbors[12] == [8, 13]
            assert pixel_neighbors[13] == [9, 12, 14]
            assert pixel_neighbors[14] == [10, 13, 15]
            assert pixel_neighbors[15] == [11, 14]


class TestVoronoi:
    class TestVoronoiGrid:

        def test__points_in_x_cross_shape__sets_up_diamond_voronoi_vertices(self):
            # 5 points in the shape of the face of a 5 on a die - makes a diamond Voronoi diagram

            points = np.array([[-1.0, 1.0], [1.0, 1.0],
                               [0.0, 0.0],
                               [-1.0, -1.0], [1.0, -1.0]])

            pix = pixelizations.Voronoi(pixels=5)
            voronoi = pix.voronoi_from_pixel_centers(points)

            voronoi.vertices = list(map(lambda x: list(x), voronoi.vertices))

            assert [0, 1.] in voronoi.vertices
            assert [-1., 0.] in voronoi.vertices
            assert [1., 0.] in voronoi.vertices
            assert [0., -1.] in voronoi.vertices

        def test__9_points_in_square___sets_up_square_of_voronoi_vertices(self):
            # 9 points in a square - makes a square (this is the example int he scipy documentaiton page)

            points = np.array([[2.0, 0.0], [2.0, 1.0], [2.0, 2.0],
                               [1.0, 0.0], [1.0, 1.0], [1.0, 2.0],
                               [0.0, 0.0], [0.0, 1.0], [0.0, 2.0]])

            pix = pixelizations.Voronoi(pixels=9)
            voronoi = pix.voronoi_from_pixel_centers(points)

            # ridge points is a numpy array for speed, but convert to list for the comparisons below so we can use in
            # to look for each list

            voronoi.vertices = list(map(lambda x: list(x), voronoi.vertices))

            assert [0.5, 1.5] in voronoi.vertices
            assert [1.5, 0.5] in voronoi.vertices
            assert [0.5, 0.5] in voronoi.vertices
            assert [1.5, 1.5] in voronoi.vertices

        def test__points_in_x_cross_shape__sets_up_pairs_of_voronoi_cells(self):
            # 5 points in the shape of the face of a 5 on a die - makes a diamond Voronoi diagram

            points = np.array([[1.0, -1.0], [1.0, 1.0],
                                      [0.0, 0.0],
                               [-1.0, -1.0], [-1.0, 1.0]])

            pix = pixelizations.Voronoi(pixels=5)
            voronoi = pix.voronoi_from_pixel_centers(points)

            # ridge points is a numpy array for speed, but convert to list for the comparisons below so we can use in
            # to look for each list

            voronoi.ridge_points = list(map(lambda x: list(x), voronoi.ridge_points))

            assert len(voronoi.ridge_points) == 8

            assert [2, 0] in voronoi.ridge_points or [0, 2] in voronoi.ridge_points
            assert [2, 1] in voronoi.ridge_points or [1, 2] in voronoi.ridge_points
            assert [2, 3] in voronoi.ridge_points or [3, 2] in voronoi.ridge_points
            assert [2, 4] in voronoi.ridge_points or [4, 2] in voronoi.ridge_points
            assert [0, 1] in voronoi.ridge_points or [1, 0] in voronoi.ridge_points
            assert [0.3] in voronoi.ridge_points or [3, 0] in voronoi.ridge_points
            assert [3, 4] in voronoi.ridge_points or [4, 3] in voronoi.ridge_points
            assert [4, 1] in voronoi.ridge_points or [1, 4] in voronoi.ridge_points

        def test__9_points_in_square___sets_up_pairs_of_voronoi_cells(self):
            # 9 points in a square - makes a square (this is the example int he scipy documentaiton page)

            points = np.array([[2.0, 0.0], [2.0, 1.0], [2.0, 2.0],
                               [1.0, 0.0], [1.0, 1.0], [1.0, 2.0],
                               [0.0, 0.0], [0.0, 1.0], [0.0, 2.0]])

            pix = pixelizations.Voronoi(pixels=9)
            voronoi = pix.voronoi_from_pixel_centers(points)

            # ridge points is a numpy array for speed, but convert to list for the comparisons below so we can use in
            # to look for each list

            voronoi.ridge_points = list(map(lambda x: list(x), voronoi.ridge_points))

            assert len(voronoi.ridge_points) == 12

            assert [0, 1] in voronoi.ridge_points or [1, 0] in voronoi.ridge_points
            assert [1, 2] in voronoi.ridge_points or [2, 1] in voronoi.ridge_points
            assert [3, 4] in voronoi.ridge_points or [4, 3] in voronoi.ridge_points
            assert [4, 5] in voronoi.ridge_points or [5, 4] in voronoi.ridge_points
            assert [6, 7] in voronoi.ridge_points or [7, 6] in voronoi.ridge_points
            assert [7, 8] in voronoi.ridge_points or [8, 7] in voronoi.ridge_points

            assert [0, 3] in voronoi.ridge_points or [3, 0] in voronoi.ridge_points
            assert [1, 4] in voronoi.ridge_points or [4, 1] in voronoi.ridge_points
            assert [4, 7] in voronoi.ridge_points or [7, 4] in voronoi.ridge_points
            assert [2, 5] in voronoi.ridge_points or [5, 2] in voronoi.ridge_points
            assert [5, 8] in voronoi.ridge_points or [8, 5] in voronoi.ridge_points
            assert [3, 6] in voronoi.ridge_points or [6, 3] in voronoi.ridge_points

    class TestNeighbors:

        def test__points_in_x_cross_shape__neighbors_of_each_pixel_correct(self):
            # 5 points in the shape of the face of a 5 on a die - makes a diamond Voronoi diagram

            points = np.array([[1.0, -1.0], [1.0, 1.0],
                                     [0.0, 0.0],
                               [-1.0, -1.0], [-1.0, 1.0]])

            pix = pixelizations.Voronoi(pixels=5)
            voronoi = pix.voronoi_from_pixel_centers(points)
            neighbors = pix.neighbors_from_pixelization(voronoi.ridge_points)

            assert set(neighbors[0]) == {2, 1, 3}
            assert set(neighbors[1]) == {2, 0, 4}
            assert set(neighbors[2]) == {0, 1, 3, 4}
            assert set(neighbors[3]) == {2, 0, 4}
            assert set(neighbors[4]) == {2, 1, 3}

        def test__9_points_in_square___neighbors_of_each_pixel_correct(self):
            # 9 points in a square - makes a square (this is the example int he scipy documentaiton page)

            points = np.array([[2.0, 0.0], [2.0, 1.0], [2.0, 2.0],
                               [1.0, 0.0], [1.0, 1.0], [1.0, 2.0],
                               [0.0, 0.0], [0.0, 1.0], [0.0, 2.0]])

            pix = pixelizations.Voronoi(pixels=9)
            voronoi = pix.voronoi_from_pixel_centers(points)
            neighbors = pix.neighbors_from_pixelization(voronoi.ridge_points)

            assert set(neighbors[0]) == {1, 3}
            assert set(neighbors[1]) == {0, 2, 4}
            assert set(neighbors[2]) == {1, 5}
            assert set(neighbors[3]) == {0, 4, 6}
            assert set(neighbors[4]) == {1, 3, 5, 7}
            assert set(neighbors[5]) == {2, 4, 8}
            assert set(neighbors[6]) == {3, 7}
            assert set(neighbors[7]) == {4, 6, 8}
            assert set(neighbors[8]) == {5, 7}


class TestAmorphous:
    class TestKMeans:

        def test__simple_points__sets_up_two_clusters(self):

            cluster_grid = np.array([[1.99, 0.99], [2.0, 1.0], [2.01, 1.01],
                                     [0.99, 1.99], [1.0, 2.0], [1.01, 2.01]])

            pix = pixelizations.Amorphous(pixels=2)

            pixel_centers, pix_to_image = pix.kmeans_cluster(cluster_grid)

            assert [2.0, 2.0] in pixel_centers
            assert [1.0, 1.0] in pixel_centers

            assert list(pix_to_image).count(0) == 3
            assert list(pix_to_image).count(1) == 3

        def test__simple_points__sets_up_three_clusters(self):
            cluster_grid = np.array([[1.99, -0.99], [2.0, -1.0], [2.01, -1.01],
                                     [0.99, 0.99], [1.0, 1.0], [1.01, 1.01],
                                     [-0.99, 1.99], [-1.0, 2.0], [-1.01, 2.01]])

            pix = pixelizations.Amorphous(pixels=3)

            pixel_centers, pix_to_image = pix.kmeans_cluster(cluster_grid)

            assert [2.0, 2.0] in pixel_centers
            assert [1.0, 1.0] in pixel_centers
            assert [-1.0, -1.0] in pixel_centers

            assert list(pix_to_image).count(0) == 3
            assert list(pix_to_image).count(1) == 3
            assert list(pix_to_image).count(2) == 3

        def test__simple_points__sets_up_three_clusters_more_points_in_third_cluster(self):
            cluster_grid = np.array([[-0.99, -0.99], [-1.0, -1.0], [-1.01, -1.01],

                                     [0.99, 0.99], [1.0, 1.0], [1.01, 1.01],
                                     [0.99, 0.99], [1.0, 1.0], [1.01, 1.01],

                                     [1.99, 1.99], [2.0, 2.0], [2.01, 2.01],
                                     [1.99, 1.99], [2.0, 2.0], [2.01, 2.01],
                                     [1.99, 1.99], [2.0, 2.0], [2.01, 2.01],
                                     [1.99, 1.99], [2.0, 2.0], [2.01, 2.01]])

            pix = pixelizations.Amorphous(pixels=3)

            pixel_centers, pix_to_image = pix.kmeans_cluster(cluster_grid)

            pixel_centers = list(map(lambda x: pytest.approx(list(x), 1e-3), pixel_centers))

            assert [2.0, 2.0] in pixel_centers
            assert [1.0, 1.0] in pixel_centers
            assert [-1.0, -1.0] in pixel_centers

            assert list(pix_to_image).count(0) == 3 or 6 or 12
            assert list(pix_to_image).count(1) == 3 or 6 or 12
            assert list(pix_to_image).count(2) == 3 or 6 or 12

            assert list(pix_to_image).count(0) != list(pix_to_image).count(1) != list(pix_to_image).count(2)


class TestPixelizationGrid:
    class TestCoordinateGridWithinAnnulus:

        def test__shape_3x3__circle_radius_15__all_9_pixels_in_grid_with_correct_coordinates(self):
            pixelization_grid = pixelizations.PixelizationGrid(shape=(3, 3))

            coordinate_grid = pixelization_grid.coordinate_grid_within_annulus(inner_radius=0.0, outer_radius=1.5)

            assert (coordinate_grid == np.array([[1.0, -1.0], [1.0, 0.0], [1.0, 1.0],
                                                 [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                                 [-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0]])).all()

        def test__shape_3x3__circle_radius_3__all_9_pixels_in_grid_with_correct_coordinates(self):
            pixelization_grid = pixelizations.PixelizationGrid(shape=(3, 3))

            coordinate_grid = pixelization_grid.coordinate_grid_within_annulus(inner_radius=0.0, outer_radius=3)

            assert (coordinate_grid == np.array([[2.0, -2.0], [2.0, 0.0], [2.0, 2.0],
                                                 [0.0, -2.0], [0.0, 0.0], [0.0, 2.0],
                                                 [-2.0, -2.0], [-2.0, 0.0], [-2.0, 2.0]])).all()

        def test__shape_3x2__circle_radius_15__all_6_pixels_in_grid_with_correct_coordinates(self):
            pixelization_grid = pixelizations.PixelizationGrid(shape=(3, 2))

            coordinate_grid = pixelization_grid.coordinate_grid_within_annulus(inner_radius=0.0, outer_radius=1.5)

            assert (coordinate_grid == np.array([[1.0, -0.75], [1.0, 0.75],
                                                 [0.0, -0.75], [0.0, 0.75],
                                                 [-1.0, -0.75], [-1.0, 0.75]])).all()

        def test__shape_2x3__circle_radius_15__all_6_pixels_in_grid_with_correct_coordinates(self):
            pixelization_grid = pixelizations.PixelizationGrid(shape=(2, 3))

            coordinate_grid = pixelization_grid.coordinate_grid_within_annulus(inner_radius=0.0, outer_radius=1.5)

            assert (coordinate_grid == np.array([[0.75, -1.0], [0.75, 0.0], [0.75, 1.0],
                                                 [-0.75, -1.0], [-0.75, 0.0], [-0.75, 1.0]])).all()



"""The base command."""

import os
from functools import wraps

current_directory = os.getcwd()


def prepend_working_directory(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        simple_path = func(*args, **kwargs)
        if not simple_path.startswith("/"):
            simple_path = "{}/{}".format(current_directory, simple_path)
        return simple_path

    return wrapper


class Base(object):
    """A base command."""

    def __init__(self, options, *args, **kwargs):
        self.options = options
        self.args = args
        self.kwargs = kwargs

    def run(self):
        raise NotImplementedError('You must implement the run() method yourself!')



import numpy as np

from autolens.imaging import scaled_array
from autolens.imaging import mask as msk


class GalaxyData(scaled_array.ScaledSquarePixelArray):

    def __new__(cls, array, mask, noise_map, sub_grid_size=2):
        return np.array(mask.map_2d_array_to_masked_1d_array(array)).view(cls)

    def __init__(self, array, noise_map, mask, sub_grid_size=2):
        """
        The lensing _data is the collection of data (images, noise-maps, PSF), a mask, grids, convolvers and other \
        utilities that are used for modeling and fitting an _data of a strong lens.

        Whilst the _data data is initially loaded in 2D, for the lensing _data the masked-_data (and noise-maps) \
        are reduced to 1D arrays for faster calculations.

        Parameters
        ----------
        array : scaled_array.ScaledSquarePixelArray
            The original _data data in 2D.
        mask: msk.Mask
            The 2D mask that is applied to the _data.
        sub_grid_size : int
            The size of the sub-grid used for each lensing SubGrid. E.g. a value of 2 grids each _data-pixel on a 2x2 \
            sub-grid.
        image_psf_shape : (int, int)
            The shape of the PSF used for convolving model images generated using analytic light profiles. A smaller \
            shape will trim the PSF relative to the input _data PSF, giving a faster analysis run-time.
        """
        super().__init__(array=array, pixel_scale=array.pixel_scale)

        self.array = array
        self.noise_map = mask.map_2d_array_to_masked_1d_array(array_2d=noise_map)
        self.mask = mask
        self.sub_grid_size = sub_grid_size

        self.grids = msk.ImagingGrids.grids_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                     sub_grid_size=sub_grid_size, psf_shape=(1,1))

        self.unmasked_grids = msk.ImagingGrids.padded_grids_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                              sub_grid_size=sub_grid_size, psf_shape=(1,1))

    def __array_finalize__(self, obj):
        super(GalaxyData, self).__array_finalize__(obj)
        if isinstance(obj, GalaxyData):
            self.array = obj.array
            self.mask = obj.mask
            self.sub_grid_size = obj.sub_grid_size
            self.grids = obj.grids
            self.unmasked_grids = obj.unmasked_grids


import numba
import numpy as np
import pytest
from profiling import profiling_data
from profiling import tools

from profiles import light_profiles


class EllipticalSersic(light_profiles.EllipticalLightProfile):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6,
                 sersic_index=4.0, settings=light_profiles.LightProfileSettings()):
        """

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        intensity : float
            Overall intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The circular radius containing half the light of this model_mapper
        sersic_index : Int
            The concentration of the light profiles
        """
        super(EllipticalSersic, self).__init__(centre, axis_ratio, phi, settings)
        self.intensity = intensity
        self.effective_radius = effective_radius
        self.sersic_index = sersic_index

    @property
    def sersic_constant(self):
        """

        Returns
        -------
        sersic_constant: float
            A parameter, derived from sersic_index, that ensures that effective_radius always contains 50% of the light.
        """
        return (2 * self.sersic_index) - (1. / 3.) + (4. / (405. * self.sersic_index)) + (
                46. / (25515. * self.sersic_index ** 2)) + (131. / (1148175. * self.sersic_index ** 3)) - (
                       2194697. / (30690717750. * self.sersic_index ** 4))

    def intensity_from_grid(self, grid):
        return self.intensity_at_grid_radii(self.grid_to_eccentric_radii(grid))

    def grid_to_eccentric_radii(self, grid):
        return np.multiply(np.sqrt(self.axis_ratio),
                           np.sqrt(np.add(np.square(grid[:, 0]),
                                          np.square(np.divide(grid[:, 1], self.axis_ratio))))).view(np.ndarray)

    def intensity_at_grid_radii(self, grid_radii):
        return np.multiply(self.intensity, np.exp(
            np.multiply(-self.sersic_constant,
                        np.add(np.power(np.divide(grid_radii, self.effective_radius), 1. / self.sersic_index), -1))))

    def intensity_from_grid_jitted(self, grid):
        return self.intensity_from_grid_jit(grid, self.axis_ratio, self.intensity, self.effective_radius,
                                            self.sersic_index,
                                            self.sersic_constant)

    @staticmethod
    @numba.jit(nopython=True)
    def intensity_from_grid_jit(grid, axis_ratio, intensity, effective_radius, sersic_index, sersic_constant):
        intensities = np.zeros(grid.shape[0])

        for i in range(grid.shape[0]):
            intensities[i] = intensity * np.exp(
                -sersic_constant * (((np.multiply(np.sqrt(axis_ratio), np.sqrt(np.add(np.square(grid[i, 0]),
                                                                                      np.square(np.divide(grid[i, 1],
                                                                                                          axis_ratio))))) / effective_radius) **
                                     (1. / sersic_index)) - 1))

        return intensities


sub_grid_size = 4

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size)
ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=sub_grid_size)

sersic = EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, intensity=0.1,
                          effective_radius=0.8, sersic_index=4.0)

assert (sersic.intensity_from_grid(grid=lsst.coords.sub_grid_coords) ==
        pytest.approx(sersic.intensity_from_grid_jitted(grid=lsst.coords.sub_grid_coords), 1e-4))


@tools.tick_toc_x10
def lsst_solution():
    sersic.intensity_from_grid_jitted(grid=lsst.coords.sub_grid_coords)


@tools.tick_toc_x10
def euclid_solution():
    sersic.intensity_from_grid_jitted(grid=euclid.coords.sub_grid_coords)


@tools.tick_toc_x10
def hst_solution():
    sersic.intensity_from_grid_jitted(grid=hst.coords.sub_grid_coords)


@tools.tick_toc_x10
def hst_up_solution():
    sersic.intensity_from_grid_jitted(grid=hst_up.coords.sub_grid_coords)


@tools.tick_toc_x10
def ao_solution():
    sersic.intensity_from_grid_jitted(grid=ao.coords.sub_grid_coords)


if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()



import numpy as np

from autolens import exc
from autolens.inversion import inversions
from autolens.fitting import fitting
from autolens.lensing import lensing_image as li
from autolens.lensing import ray_tracing

minimum_value_profile = 0.1


def fit_lensing_image_with_tracer(lensing_image, tracer, padded_tracer=None):
    """Fit.

    Parameters
    -----------
    lensing_image : li.LensingImage
        Lensing _data
    tracer : ray_tracing.AbstractTracer
        tracer
    """

    if tracer.has_light_profile and not tracer.has_pixelization:

        if not tracer.has_hyper_galaxy:
            return LensingProfileFit(lensing_image=lensing_image, tracer=tracer, padded_tracer=padded_tracer)
        elif tracer.has_hyper_galaxy:
            return HyperLensingProfileFit(lensing_hyper_image=lensing_image, tracer=tracer, padded_tracer=padded_tracer)

    elif not tracer.has_light_profile and tracer.has_pixelization:

        if not tracer.has_hyper_galaxy:
            return LensingInversionFit(lensing_image=lensing_image, tracer=tracer)
        elif tracer.has_hyper_galaxy:
            return HyperLensingInversionFit(lensing_hyper_image=lensing_image, tracer=tracer)

    elif tracer.has_light_profile and tracer.has_pixelization:

        if not tracer.has_hyper_galaxy:
            return LensingProfileInversionFit(lensing_image=lensing_image, tracer=tracer, padded_tracer=padded_tracer)
        elif tracer.has_hyper_galaxy:
            return HyperLensingProfileInversionFit(lensing_hyper_image=lensing_image, tracer=tracer,
                                                   padded_tracer=padded_tracer)

    else:

        raise exc.FittingException('The fit routine did not call a Fit class - check the '
                                   'properties of the tracer')


def fast_likelihood_from_lensing_image_and_tracer(lensing_image, tracer):
    if tracer.has_light_profile and not tracer.has_pixelization:

        if not tracer.has_hyper_galaxy:
            return LensingProfileFit.fast_likelihood(lensing_image=lensing_image, tracer=tracer)
        elif tracer.has_hyper_galaxy:
            return HyperLensingProfileFit.fast_scaled_likelihood(lensing_hyper_image=lensing_image, tracer=tracer)

    elif not tracer.has_light_profile and tracer.has_pixelization:

        if not tracer.has_hyper_galaxy:
            return LensingInversionFit.fast_evidence(lensing_image=lensing_image, tracer=tracer)
        elif tracer.has_hyper_galaxy:
            return HyperLensingInversionFit.fast_scaled_evidence(lensing_image=lensing_image, tracer=tracer)

    elif tracer.has_light_profile and tracer.has_pixelization:

        if not tracer.has_hyper_galaxy:
            return LensingProfileInversionFit.fast_evidence(lensing_image=lensing_image, tracer=tracer)
        elif tracer.has_hyper_galaxy:
            return HyperLensingProfileInversionFit.fast_scaled_evidence(lensing_image=lensing_image, tracer=tracer)

    else:

        raise exc.FittingException('The fast likelihood routine did not call a likelihood functions - check the '
                                   'properties of the tracer')


class AbstractLensingFit(object):

    def __init__(self, lensing_image, tracer, padded_tracer=None):

        self.lensing_image = lensing_image
        self.tracer = tracer
        self.padded_tracer = padded_tracer

    @property
    def total_inversions(self):
        return len(self.tracer.mappers_of_planes)


class LensingProfileFit(fitting.AbstractProfileFit, AbstractLensingFit):

    def __init__(self, lensing_image, tracer, padded_tracer=None):
        """
        Class to evaluate the fit between a model described by a tracer and an actual lensing_image.

        Parameters
        ----------
        lensing_image: li.LensingImage
            An lensing_image that has been masked for efficiency
        tracer: ray_tracing.AbstractTracer
            An object describing the model
        """
        AbstractLensingFit.__init__(self=self, lensing_image=lensing_image, tracer=tracer, padded_tracer=padded_tracer)

        super(LensingProfileFit, self).__init__(fitting_image=lensing_image, _image=tracer._image_plane_image,
                                                _blurring_image=tracer._image_plane_blurring_image)

    @classmethod
    def fast_likelihood(cls, lensing_image, tracer):
        convolve_image = lensing_image.convolver_image.convolve_image
        _model_image = convolve_image(tracer._image_plane_image, tracer._image_plane_blurring_image)
        _residuals = fitting.residuals_from_data_and_model(lensing_image[:], _model_image)
        _chi_squareds = fitting.chi_squareds_from_residuals_and_noise(_residuals, lensing_image.noise_map)
        chi_squared_term = fitting.chi_squared_term_from_chi_squareds(_chi_squareds)
        noise_term = fitting.noise_term_from_noise_map(lensing_image.image.noise_map)
        return fitting.likelihood_from_chi_squared_and_noise_terms(chi_squared_term, noise_term)

    @property
    def model_images_of_planes(self):

        _model_images_of_planes = list(map(lambda plane:
                                  self.convolver_image.convolve_image(plane._image_plane_image,
                                                                      plane._image_plane_blurring_image),
                                  self.tracer.all_planes))

        return list(map(lambda image : None if not image.any() else self.map_to_scaled_array(image),
                         _model_images_of_planes))

    @property
    def unmasked_model_profile_image(self):
        if self.padded_tracer is None:
            return None
        elif self.padded_tracer is not None:
            return fitting.unmasked_model_image_from_fitting_image(self.lensing_image,
                                                                   self.padded_tracer._image_plane_image)

    @property
    def unmasked_model_profile_images_of_galaxies(self):
        if self.padded_tracer is None:
            return None
        elif self.padded_tracer is not None:
            return unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(self.lensing_image,
                                                                                   self.padded_tracer)


class HyperLensingProfileFit(LensingProfileFit, fitting.AbstractHyperFit):

    def __init__(self, lensing_hyper_image, tracer, padded_tracer=None):
        """
        Class to evaluate the fit between a model described by a tracer and an actual lensing_image.

        Parameters
        ----------
        lensing_hyper_image: li.LensingHyperImage
            An lensing_image that has been masked for efficiency
        tracer: ray_tracing.AbstractTracer
            An object describing the model
        """

        fitting.AbstractHyperFit.__init__(self=self, fitting_hyper_image=lensing_hyper_image,
                                          hyper_galaxies=tracer.hyper_galaxies)
        super(HyperLensingProfileFit, self).__init__(lensing_hyper_image, tracer, padded_tracer)

        self._scaled_chi_squareds = fitting.chi_squareds_from_residuals_and_noise(self._residuals,
                                                                                  self._scaled_noise_map)

    @classmethod
    def fast_scaled_likelihood(cls, lensing_hyper_image, tracer):
        _contributions = fitting.contributions_from_hyper_images_and_galaxies(lensing_hyper_image.hyper_model_image,
                         lensing_hyper_image.hyper_galaxy_images, tracer.hyper_galaxies,
                         lensing_hyper_image.hyper_minimum_values)
        _scaled_noise_map = fitting.scaled_noise_from_hyper_galaxies_and_contributions(_contributions, tracer.hyper_galaxies,
                                                                                       lensing_hyper_image.noise_map)

        convolve_image = lensing_hyper_image.convolver_image.convolve_image
        _model_image = convolve_image(tracer._image_plane_image, tracer._image_plane_blurring_image)
        _residuals = fitting.residuals_from_data_and_model(lensing_hyper_image[:], _model_image)
        _scaled_chi_squareds = fitting.chi_squareds_from_residuals_and_noise(_residuals, _scaled_noise_map)

        scaled_chi_squared_term = fitting.chi_squared_term_from_chi_squareds(_scaled_chi_squareds)
        scaled_noise_term = fitting.noise_term_from_noise_map(_scaled_noise_map)
        return fitting.likelihood_from_chi_squared_and_noise_terms(scaled_chi_squared_term, scaled_noise_term)

    @property
    def scaled_likelihood(self):
        return fitting.likelihood_from_chi_squared_and_noise_terms(self.scaled_chi_squared_term, self.scaled_noise_term)


class LensingInversionFit(fitting.AbstractInversionFit, AbstractLensingFit):

    def __init__(self, lensing_image, tracer):
        """
        Class to evaluate the fit between a model described by a tracer and an actual lensing_image.

        Parameters
        ----------
        lensing_image: li.LensingImage
            An lensing_image that has been masked for efficiency
        tracer: ray_tracing.AbstractTracer
            An object describing the model
        """

        AbstractLensingFit.__init__(self=self, lensing_image=lensing_image, tracer=tracer)

        mapper = tracer.mappers_of_planes[0]
        regularization = tracer.regularizations_of_planes[0]

        super(LensingInversionFit, self).__init__(fitting_image=lensing_image, mapper=mapper,
                                                  regularization=regularization)

    @classmethod
    def fast_evidence(cls, lensing_image, tracer):
        mapper = tracer.mappers_of_planes[0]
        regularization = tracer.regularizations_of_planes[0]
        inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(lensing_image[:], lensing_image.noise_map,
                                                                                      lensing_image.convolver_mapping_matrix,
                                                                                      mapper, regularization)
        _model_image = inversion.reconstructed_data_vector
        _residuals = fitting.residuals_from_data_and_model(lensing_image[:], _model_image)
        _chi_squareds = fitting.chi_squareds_from_residuals_and_noise(_residuals, lensing_image.noise_map)
        chi_squared_term = fitting.chi_squared_term_from_chi_squareds(_chi_squareds)
        noise_term = fitting.noise_term_from_noise_map(lensing_image.noise_map)
        return fitting.evidence_from_reconstruction_terms(chi_squared_term, inversion.regularization_term,
                                                  inversion.log_det_curvature_reg_matrix_term,
                                                  inversion.log_det_regularization_matrix_term, noise_term)

    @property
    def model_images_of_planes(self):
        return [None, self.map_to_scaled_array(self._model_data)]


class HyperLensingInversion(fitting.AbstractHyperFit):

    @property
    def scaled_model_image(self):
        return self.map_to_scaled_array(self._scaled_model_image)

    @property
    def scaled_residuals(self):
        return self.map_to_scaled_array(self._scaled_residuals)

    @property
    def scaled_evidence(self):
        return fitting.evidence_from_reconstruction_terms(self.scaled_chi_squared_term,
                                                  self.scaled_inversion.regularization_term,
                                                  self.scaled_inversion.log_det_curvature_reg_matrix_term,
                                                  self.scaled_inversion.log_det_regularization_matrix_term,
                                                  self.scaled_noise_term)


class HyperLensingInversionFit(LensingInversionFit, HyperLensingInversion):

    def __init__(self, lensing_hyper_image, tracer):
        """
        Class to evaluate the fit between a model described by a tracer and an actual lensing_image.

        Parameters
        ----------
        lensing_hyper_image: li.LensingHyperImage
            An lensing_image that has been masked for efficiency
        tracer: ray_tracing.AbstractTracer
            An object describing the model
        """

        fitting.AbstractHyperFit.__init__(self=self, fitting_hyper_image=lensing_hyper_image,
                                          hyper_galaxies=tracer.hyper_galaxies)
        super(HyperLensingInversionFit, self).__init__(lensing_hyper_image, tracer)

        self.scaled_inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(
            lensing_hyper_image[:], self._scaled_noise_map, lensing_hyper_image.convolver_mapping_matrix,
            self.inversion.mapper, self.inversion.regularization)

        self._scaled_model_image = self.scaled_inversion.reconstructed_data_vector
        self._scaled_residuals = fitting.residuals_from_data_and_model(self._data, self._scaled_model_image)
        self._scaled_chi_squareds = fitting.chi_squareds_from_residuals_and_noise(self._scaled_residuals,
                                                                          self._scaled_noise_map)

    @classmethod
    def fast_scaled_evidence(cls, lensing_image, tracer):
        _contributions = fitting.contributions_from_hyper_images_and_galaxies(lensing_image.hyper_model_image,
                                                                              lensing_image.hyper_galaxy_images,
                                                                      tracer.hyper_galaxies,
                                                                              lensing_image.hyper_minimum_values)
        _scaled_noise_map = fitting.scaled_noise_from_hyper_galaxies_and_contributions(_contributions, tracer.hyper_galaxies,
                                                                               lensing_image.noise_map)

        mapper = tracer.mappers_of_planes[0]
        regularization = tracer.regularizations_of_planes[0]

        scaled_inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(
            lensing_image[:], _scaled_noise_map, lensing_image.convolver_mapping_matrix, mapper, regularization)

        _scaled_residuals = fitting.residuals_from_data_and_model(lensing_image[:], scaled_inversion.reconstructed_data_vector)
        _scaled_chi_squareds = fitting.chi_squareds_from_residuals_and_noise(_scaled_residuals, _scaled_noise_map)
        scaled_chi_squared_term = fitting.chi_squared_term_from_chi_squareds(_scaled_chi_squareds)
        scaled_noise_term = fitting.noise_term_from_noise_map(_scaled_noise_map)
        return fitting.evidence_from_reconstruction_terms(scaled_chi_squared_term, scaled_inversion.regularization_term,
                                                  scaled_inversion.log_det_curvature_reg_matrix_term,
                                                  scaled_inversion.log_det_regularization_matrix_term,
                                                  scaled_noise_term)


class LensingProfileInversionFit(fitting.AbstractProfileInversionFit, AbstractLensingFit):

    def __init__(self, lensing_image, tracer, padded_tracer=None):
        """
        Class to evaluate the fit between a model described by a tracer and an actual lensing_image.

        Parameters
        ----------
        lensing_image: li.LensingImage
            An lensing_image that has been masked for efficiency
        tracer: ray_tracing.AbstractTracer
            An object describing the model
        """

        mapper = tracer.mappers_of_planes[0]
        regularization = tracer.regularizations_of_planes[0]

        AbstractLensingFit.__init__(self=self, lensing_image=lensing_image, tracer=tracer, padded_tracer=padded_tracer)
        super(LensingProfileInversionFit, self).__init__(fitting_image=lensing_image, _image=tracer._image_plane_image,
                                                         _blurring_image=tracer._image_plane_blurring_image,
                                                         mapper=mapper, regularization=regularization)

    @classmethod
    def fast_evidence(cls, lensing_image, tracer):
        convolve_image = lensing_image.convolver_image.convolve_image
        _profile_model_image = convolve_image(tracer._image_plane_image, tracer._image_plane_blurring_image)
        _profile_subtracted_image = lensing_image[:] - _profile_model_image
        mapper = tracer.mappers_of_planes[0]
        regularization = tracer.regularizations_of_planes[0]
        inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(_profile_subtracted_image,
                                                                                      lensing_image.noise_map,
                                                                                      lensing_image.convolver_mapping_matrix,
                                                                                      mapper, regularization)
        _model_image = _profile_model_image + inversion.reconstructed_data_vector
        _residuals = fitting.residuals_from_data_and_model(lensing_image[:], _model_image)
        _chi_squareds = fitting.chi_squareds_from_residuals_and_noise(_residuals, lensing_image.noise_map)
        chi_squared_term = fitting.chi_squared_term_from_chi_squareds(_chi_squareds)
        noise_term = fitting.noise_term_from_noise_map(lensing_image.noise_map)
        return fitting.evidence_from_reconstruction_terms(chi_squared_term, inversion.regularization_term,
                                                  inversion.log_det_curvature_reg_matrix_term,
                                                  inversion.log_det_regularization_matrix_term, noise_term)

    @property
    def model_images_of_planes(self):
        return [self.profile_model_image, self.inversion_model_image]


class HyperLensingProfileInversionFit(LensingProfileInversionFit, HyperLensingInversion):

    def __init__(self, lensing_hyper_image, tracer, padded_tracer=None):
        """
        Class to evaluate the fit between a model described by a tracer and an actual lensing_image.

        Parameters
        ----------
        lensing_hyper_image: li.LensingHyperImage
            An lensing_image that has been masked for efficiency
        tracer: ray_tracing.AbstractTracer
            An object describing the model
        """

        fitting.AbstractHyperFit.__init__(self=self, fitting_hyper_image=lensing_hyper_image,
                                          hyper_galaxies=tracer.hyper_galaxies)
        super(HyperLensingProfileInversionFit, self).__init__(lensing_hyper_image, tracer, padded_tracer)

        self.scaled_inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(
            self._profile_subtracted_image, self._scaled_noise_map, lensing_hyper_image.convolver_mapping_matrix,
            self.inversion.mapper, self.inversion.regularization)

        self._scaled_model_image = self._profile_model_image + self.scaled_inversion.reconstructed_data_vector
        self._scaled_residuals = fitting.residuals_from_data_and_model(self._data, self._scaled_model_image)
        self._scaled_chi_squareds = fitting.chi_squareds_from_residuals_and_noise(self._scaled_residuals,
                                                                          self._scaled_noise_map)

    @classmethod
    def fast_scaled_evidence(cls, lensing_image, tracer):
        _contributions = fitting.contributions_from_hyper_images_and_galaxies(lensing_image.hyper_model_image,
                                                                              lensing_image.hyper_galaxy_images,
                                                                              tracer.hyper_galaxies,
                                                                              lensing_image.hyper_minimum_values)
        _scaled_noise_map = fitting.scaled_noise_from_hyper_galaxies_and_contributions(_contributions, tracer.hyper_galaxies,
                                                                               lensing_image.noise_map)

        convolve_image = lensing_image.convolver_image.convolve_image
        _profile_model_image = convolve_image(tracer._image_plane_image, tracer._image_plane_blurring_image)
        _profile_subtracted_image = lensing_image[:] - _profile_model_image

        mapper = tracer.mappers_of_planes[0]
        regularization = tracer.regularizations_of_planes[0]

        scaled_inversion = inversions.inversion_from_lensing_image_mapper_and_regularization(
            _profile_subtracted_image, _scaled_noise_map, lensing_image.convolver_mapping_matrix, mapper,
            regularization)

        _scaled_model_image = _profile_model_image + scaled_inversion.reconstructed_data_vector
        _scaled_residuals = fitting.residuals_from_data_and_model(lensing_image[:], _scaled_model_image)
        _scaled_chi_squareds = fitting.chi_squareds_from_residuals_and_noise(_scaled_residuals, _scaled_noise_map)
        scaled_chi_squared_term = fitting.chi_squared_term_from_chi_squareds(_scaled_chi_squareds)
        scaled_noise_term = fitting.noise_term_from_noise_map(_scaled_noise_map)
        return fitting.evidence_from_reconstruction_terms(scaled_chi_squared_term, scaled_inversion.regularization_term,
                                                  scaled_inversion.log_det_curvature_reg_matrix_term,
                                                  scaled_inversion.log_det_regularization_matrix_term,
                                                  scaled_noise_term)


class PositionFit:

    def __init__(self, positions, noise):

        self.positions = positions
        self.noise = noise

    @property
    def chi_squareds(self):
        return np.square(np.divide(self.maximum_separations, self.noise))

    @property
    def likelihood(self):
        return -0.5 * sum(self.chi_squareds)

    def maximum_separation_within_threshold(self, threshold):
        if max(self.maximum_separations) > threshold:
            return False
        else:
            return True

    @property
    def maximum_separations(self):
        return list(map(lambda positions: self.max_separation_of_grid(positions), self.positions))

    def max_separation_of_grid(self, grid):
        rdist_max = np.zeros((grid.shape[0]))
        for i in range(grid.shape[0]):
            xdists = np.square(np.subtract(grid[i, 0], grid[:, 0]))
            ydists = np.square(np.subtract(grid[i, 1], grid[:, 1]))
            rdist_max[i] = np.max(np.add(xdists, ydists))
        return np.max(np.sqrt(rdist_max))


# TODO : The [plane_index][galaxy_index] data structure is going to be key to tracking galaxies / hyper galaxies in
# TODO : Multi-plane ray tracing. I never felt it was easy to follow using list comprehensions from ray_tracing.
# TODO : Can we make this neater?


def unmasked_model_images_of_galaxies_from_lensing_image_and_tracer(lensing_image, tracer):


    padded_model_images_of_galaxies = [[] for _ in range(len(tracer.all_planes))]

    for plane_index, plane in enumerate(tracer.all_planes):
        for galaxy_index in range(len(plane.galaxies)):

            _galaxy_image_plane_image = plane._image_plane_image_of_galaxies[galaxy_index]

            galaxy_model_image = fitting.unmasked_model_image_from_fitting_image(fitting_image=lensing_image,
                                                                   _unmasked_image=_galaxy_image_plane_image)

            padded_model_images_of_galaxies[plane_index].append(galaxy_model_image)

    return padded_model_images_of_galaxies